#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>

#define STR_SIZE 64

typedef enum { false, true } bool;

const char* filename = "master";
const char* lock_filename = "master.lock";

void print_usage(char* pname)
{
	printf("usage: %s [c|o]\n", pname);
}

void rand_str(char *dest, size_t length)
{
    char charset[] = "0123456789"
                     "abcdefghijklmnopqrstuvwxyz"
                     "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    srand(time(0));
    while (length-- > 0) {
        size_t index = (double) rand() / RAND_MAX * (sizeof charset - 1);
        *dest++ = charset[index];
    }
    *dest = '\0';
}

void statFile()
{
    struct stat fileStat;
    if(stat(filename, &fileStat) < 0)
    {
        perror("stat(master) : ");
        exit(EXIT_FAILURE);
    }

    printf("[%s] Size: %zu bytes. nlink %zu, Inode: %zx ",
            filename,
            fileStat.st_size,
            fileStat.st_nlink,
            fileStat.st_ino);

    printf(" File Permissions:");
    printf( (S_ISDIR(fileStat.st_mode)) ? "d" : "-");
    printf( (fileStat.st_mode & S_IRUSR) ? "r" : "-");
    printf( (fileStat.st_mode & S_IWUSR) ? "w" : "-");
    printf( (fileStat.st_mode & S_IXUSR) ? "x" : "-");
    printf( (fileStat.st_mode & S_IRGRP) ? "r" : "-");
    printf( (fileStat.st_mode & S_IWGRP) ? "w" : "-");
    printf( (fileStat.st_mode & S_IXGRP) ? "x" : "-");
    printf( (fileStat.st_mode & S_IROTH) ? "r" : "-");
    printf( (fileStat.st_mode & S_IWOTH) ? "w" : "-");
    printf( (fileStat.st_mode & S_IXOTH) ? "x" : "-");

    //printf("The file %s a symbolic link\n", (S_ISLNK(fileStat.st_mode)) ? "is" : "is not");
}

void readFile()
{
    printf("read file");
    errno = 0;
    int fd = open(filename,  O_RDWR|O_EXCL, 0666);
    if(-1 == fd)
    {
        printf("open(%s) failed with error [%s]\n", filename, strerror(errno));
        exit(EXIT_FAILURE);
    }

    char data[STR_SIZE];

    if(read(fd, data, STR_SIZE) < 0)
    {
        perror("read(master) :");
        exit(EXIT_FAILURE);
    }

    printf("Data read: %s\n", data);
    close(fd);
}

void commit(bool doRead, bool doStat, bool doWrite)
{
    if(doRead == true)
        readFile();

    if(doStat == true)
        statFile();

    if(doWrite == false)
        return;

    char rand_value[STR_SIZE];

	printf("commit() filename = %s, lock_filename = %s\n",
           filename, lock_filename);

    errno = 0;
    int fdlock = open(lock_filename,  O_RDWR|O_CREAT|O_EXCL, 0666);
    if(-1 == fdlock)
    {
        printf("open(%s) failed with error [%s]\n", lock_filename, strerror(errno));
        return;
    }

    errno = 0;
    int fd = open(filename,  O_RDWR|O_EXCL, 0666);
    if(-1 == fd)
    {
        printf("open(%s) failed with error [%s]\n", filename, strerror(errno));
        return;
    }

    rand_str(rand_value, STR_SIZE);
    if(write(fdlock, rand_value, STR_SIZE) != STR_SIZE)
    {
        printf("invalid number of bytes written to fdlock=%d\n", fdlock);
    }

    close(fdlock);
    if(rename(lock_filename, filename) < 0)
    {
        perror("rename(master.lock, master) :");
        exit(EXIT_FAILURE);
    }

    close(fd);
}

void optimize()
{
    int fdlock = open(lock_filename,  O_RDWR|O_CREAT|O_EXCL, 0666);
    if(-1 == fdlock)
    {
        printf("open(%s) failed with error [%s]\n", lock_filename, strerror(errno));
        return;
    }

    errno = 0;
    int fd = open(filename,  O_RDWR|O_EXCL, 0666);
    if(-1 == fd)
    {
        printf("open(%s) failed with error [%s]\n", filename, strerror(errno));
        return;
    }

    // I am doing some work gc
    //sleep(2);
    //unlink(lock_filename);
    close(fd);
    close(fdlock);
    unlink(lock_filename);

    printf("Optimize done\n");
}

int main(int argc, char *argv[])
{
    if ( argc != 2 )
    {
    	print_usage(argv[0]);
        return EXIT_FAILURE;
    }

    if(strcmp(argv[1], "c") == 0)
    	commit(true /*doRead*/, true /*doStat*/, true /*doWrite*/);
    else if(strcmp(argv[1], "o") == 0)
    	commit(false /*doRead*/, true /*doStat*/, true /*doWrite*/);
    else
    	print_usage(argv[0]);

   	return EXIT_SUCCESS;
}

